import {describe, it, beforeEach, afterEach, expect, vi} from 'vitest'
import {useCommunityStore} from '@/stores/community'
import { publicAxios } from '@/services/authService'
import type { Axios } from 'axios'
import type { Mocked } from 'vitest'
import { createPinia, setActivePinia } from 'pinia'


vi.mock(
    '@/services/authService',
    () => ({
      publicAxios: {
        get: vi.fn(),
        post: vi.fn()
      }
    })
)

const mockedAxios: Mocked<Axios> = vi.mocked(publicAxios, true)

describe('store/community', () => {

  setActivePinia(createPinia())
  const store = useCommunityStore()

  describe('actions', () => {
    beforeEach(() => {
      store.$reset()
    })
    afterEach(() => {
      vi.restoreAllMocks()
    })
    describe('details', () => {
      const objectId = '2378-e328'
      // beforeEach(() => {
      //   getStub.resolves({data: { id: objectId, sets: [] }})
      // })
      describe('loadProject', () => {
        beforeEach(async () => {
          mockedAxios.get.mockResolvedValueOnce({data: { id: objectId, sets: [] }})
          await store.loadProject(objectId)
        })
        it('should get project detail from api', () => {
          expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `projects/${objectId}/`)
        })
        it('should load project with received data', () => {
          expect(store.currentProject.id).toEqual(objectId)
        })
      })
      describe('loadSet', () => {
        beforeEach(async () => {
          mockedAxios.get.mockResolvedValueOnce({data: { id: objectId, items: [] }})
          await store.loadSet({setId: objectId})
        })
        it('should get set detail', () => {
          expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `sets/${objectId}/`)
        })
        it('should load set with received data', () => {
          expect(store.currentSet.id).toEqual(objectId)
        })
      })
      describe('loadItem', () => {
        describe('without template', () => {
          beforeEach(async () => {
            mockedAxios.get.mockResolvedValueOnce({data: { id: objectId}})
            await store.loadItem(objectId, false)
          })
          it('should get item detail', async () => {
            expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `items/${objectId}/`)
          })
          it('should load item with received data', async () => {
            expect(store.currentItem?.id).toEqual(objectId)
          })
        })
        it('should load template with received templateId if not specify', async () => {
          const templateId = "xx"
          store.loadTemplate = vi.fn()
          mockedAxios.get.mockResolvedValueOnce({data: { id: objectId, set: { templateId }}})
          await store.loadItem(objectId)
          expect(store.loadTemplate).toHaveBeenNthCalledWith(1, templateId)
        })
      })
    })
    describe('lists', () => {
      beforeEach(() => {
        mockedAxios.get.mockResolvedValueOnce({
          data: {
            results: [{}, {}],
            count: 2,
            nextPage: null
          }
        })
      })
      describe('loadProjectList', () => {
        beforeEach(async () => {
          await store.loadProjectList({})
        })
        it('should get project list ', () => {
          expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `projects/?page=1&page_size=10`)
        })
        it('should load project list with received data', () => {
          expect(store.projectList.pageContent).toHaveLength(2)
        })
      })
      describe('loadSetList', () => {
        const projectId = '16c5'
        beforeEach(async () => {
          await store.loadSetList(projectId)
        })
        it('should get set list ', () => {
          expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `sets/?project_id=${projectId}&page=1&page_size=10`)
        })
        it('should load set list with received data', () => {
          expect(store.setList.pageContent).toHaveLength(2)
        })
      })
      describe('loadItemList', () => {
        const setId = '16c5'
        beforeEach(async () => {
          await store.loadItemList({setId})
        })
        it('should get item list ', () => {
          expect(mockedAxios.get).toHaveBeenNthCalledWith(1, `items/?set_id=${setId}&page=1&page_size=10`)
        })
        it('should load item list with received data', () => {
          expect(store.itemList.pageContent).toHaveLength(2)
        })
      })
    })

  })
})
