import {describe, it, beforeEach, afterEach, expect, vi} from 'vitest'
import { createPinia, setActivePinia } from 'pinia'
import {useContribStore} from '@/stores/contrib'
import { authAxios } from '@/services/authService'
import { notifyError, notifySuccess} from '@/services/alertService'
import { t } from '@/plugins/i18n'
import { getUrlFromOrigin, setNewLocation } from '@/stores/helpers'
import { initUserInfo } from '../../../src/stores/contrib'
import router from '@/router'
import { createApp } from 'vue'
import { useGroupStore } from '@/stores/group'
import type { Router } from 'vue-router'
import type { Mocked } from 'vitest'
import type { UserInvitation, UserProfile } from '#/user'
import type { Authority } from '@/services/authService'


const app = createApp({})

vi.mock(
    '@/stores/helpers',
    () => ({
      getUrlFromOrigin: vi.fn(),
      setNewLocation: vi.fn()
    })
)
const mockedGetUrl = vi.mocked(getUrlFromOrigin, true)
const mockedSetUrl = vi.mocked(setNewLocation, true)


vi.mock(
    '@/services/alertService',
    () => ({
      notifyError: vi.fn(),
      notifySuccess: vi.fn()
    })
)

vi.mock(
    '@/services/authService',
    () => ({
      authAxios: {
        get: vi.fn(),
        patch: vi.fn(),
        post: vi.fn(),
        logIn: vi.fn(),
        logOut: vi.fn()
      }
    })
)
const mockedAxios: Mocked<Authority> = vi.mocked(authAxios, true)

vi.mock(
  '@/router',
  () => ({
    default: {
      replace: vi.fn()
    }
  })
)
const mockedRouter: Mocked<Router> = vi.mocked(router, true)

// vi.mock(
//   '@/services/usersService',
//   () => ({
//     getProfile: vi.fn(),
//     patchEmail: vi.fn(),
//     patchPassword: vi.fn(),
//     patchProfile: vi.fn(),
//     postExternalUserInvitation: vi.fn(),
//     postInternalUserInvitation: vi.fn(),
//     postInvitedUser: vi.fn(),
//     postResetPassword: vi.fn(),
//   })
// )
// const mockedUsersService = vi.mocked(usersService, true)

describe('store/contrib', () => {
  const pinia = createPinia()
  pinia.use(({ store }) => {
    store.router = mockedRouter
  })
  app.use(pinia)
  setActivePinia(pinia)
  const store = useContribStore()

  describe('action', () => {
    const axiosError = new Error('api call failed')
    const testError = (apiMethod: 'get' | 'patch' | 'post' | 'logIn', message: string, storeAction: Function, ...params: any[]) =>
        async () => {
          mockedAxios[apiMethod].mockRejectedValueOnce(axiosError)
          await storeAction(...params)
          expect(notifyError).toHaveBeenCalledOnce()
          expect(notifyError).toHaveBeenLastCalledWith(
              message,
              axiosError
          )
        }
    const testSuccess = (message: string) => {
      expect(notifySuccess).toHaveBeenCalledOnce()
      expect(notifySuccess).toHaveBeenLastCalledWith(message)
    }

    const appOrigin = 'https://test.url'

    const email = 'me@mail.org',
        firstName = 'me',
        lastName = 'be',
        userId = 'e579a-e87',
        username = 'CAS/username',
        password = 'cryptic phrase',
        projectId = '456'

    const userInvitation: UserInvitation = {
      invitationId: '23',
      projectId,
      email,
      firstName,
      lastName,
      newPassword: password,
      newPassword2: password
    }
    const userProfile: UserProfile = {
      id: userId,
      baseUsername: username.split('/')[1],
      username,
      firstName,
      lastName,
      email,
      isProjectCreator: false,
      isExternal: false,
      isAdmin: false,
      info: {
        bio: 'on vacation',
        researchLabs: [],
        institution: '',
        otherAffiliations: []
      },
      settings: {property: 'x'}
    }

    beforeEach(() => {
      store.$reset()
      mockedGetUrl.mockImplementation((path) => `${appOrigin}${path}`)
    })
    afterEach(() => {
      vi.restoreAllMocks()
    })
    describe('createInvitedUser', () => {
      describe('success scenario', () => {
        let loginBackUp: typeof store.logIn
        beforeEach(async () => {
          loginBackUp = store.logIn
          store.logIn = vi.fn().mockResolvedValueOnce({})
          mockedAxios.post.mockResolvedValueOnce({data: userProfile})
          await store.createInvitedUser(userInvitation)
        })
        afterEach(() => {
          store.logIn = loginBackUp
        })
        it('should post its data to api', () => {
          expect(mockedAxios.post).toHaveBeenCalledOnce()
          expect(mockedAxios.post).toHaveBeenCalledWith('users/invite/signup/', userInvitation)
        })
        it('should set user profile with received data', () => {
          expect(store.profile).toEqual(userProfile)
        })
        it('should should log user in', () => {
          expect(store.logIn).toHaveBeenCalledOnce()
          const expectedRoute = {
            name: 'projectDetail',
            params: { projectId: projectId },
            query: { welcome: 'true' }
          }
          const loginCredentials = {
            password: userInvitation.newPassword,
            username: userInvitation.email
          }
          expect(store.logIn).toHaveBeenLastCalledWith(
              expectedRoute,
              loginCredentials
          )
        })
      })
      it('should notify service error', testError(
          'post',
          t('notify.error.createUser'),
          store.createInvitedUser, userInvitation
      ))
    })
    describe('inviteExternalUser', () => {
      it('should post its data to api', async () => {
        mockedAxios.post.mockResolvedValueOnce({data: null})
        const testUrl = `${appOrigin}/invitation`
        await store.inviteExternalUser(projectId, email)
        expect(mockedAxios.post).toHaveBeenCalledOnce()
        expect(mockedAxios.post).toHaveBeenCalledWith('users/invite/', {url: testUrl, email, projectId})
      })
      it('should notify its success', async () => {
        mockedAxios.post.mockResolvedValueOnce({data: null})
        await store.inviteExternalUser(projectId, email)
        testSuccess(t('mail.success', {email}))
      })
      it('should notify service error', testError(
          'post',
          t('mail.failure', { email }),
          store.inviteExternalUser, projectId, email
      ))
    })
    describe('inviteInternalUser', () => {
      const user = { id: '564e-768a', name: '', username: username }
      const groupStore = useGroupStore()
      describe('success scenario', () => {
        beforeEach(() => {
          groupStore.addProjectMember = vi.fn()
        })
        it('should post its data to api', async () => {
          mockedAxios.post.mockResolvedValueOnce({ data: { email: '', user } })
          await store.inviteInternalUser(projectId, username)
          const projectUrl = `${appOrigin}/projects/${projectId}/`
          expect(mockedAxios.post, 'call count').toHaveBeenCalledOnce()
          expect(mockedAxios.post, 'payload check').toHaveBeenCalledWith(
              'users/invite/internal/',
              {username, projectId, url: projectUrl}
          )
        })
        it('should add user to project members', async () => {
          mockedAxios.post.mockResolvedValueOnce({ data: { email: '', user } })
          await store.inviteInternalUser(projectId, username)
          expect(groupStore.addProjectMember).toHaveBeenNthCalledWith(1, user)
        })
        it('should notify if an email as been sent', async () => {
          mockedAxios.post.mockResolvedValueOnce({ data: { email: 'sent', user } })
          await store.inviteInternalUser(projectId, username)
          testSuccess(t('group.members.mail.sent'))
        })
        it('should notify if email could not be sent', async () => {
          mockedAxios.post.mockResolvedValueOnce({ data: { email: 'error', user } })
          await store.inviteInternalUser(projectId, username)
          testSuccess(t('group.members.mail.error'))
        })
      })
      it('should notify service error', testError(
          'post',
          t('invite.failure', {username}),
          store.inviteInternalUser, projectId, username
      ))
    })
    describe('loadProfile', () => {
      const defaultUserInfo = initUserInfo()
      it('should should call user profile api',  async () => {
        mockedAxios.get.mockResolvedValueOnce({data: userProfile})
        await store.loadProfile()
        expect(mockedAxios.get).toHaveBeenNthCalledWith(1, 'users/profile/')
      })
      it('should set user profile with received data', async () => {
        mockedAxios.get.mockResolvedValueOnce({data: userProfile})
        await store.loadProfile()
        expect(store.profile).toEqual(userProfile)
      })
      it('should set user profile info to default init value', async () => {
        mockedAxios.get.mockResolvedValueOnce({
          data: {...userProfile, info: {}}
        })
        await store.loadProfile()
        expect(store.profile.info).toEqual(defaultUserInfo)
      })
      it('should set missing user profile info to default init value', async () => {
        mockedAxios.get.mockResolvedValueOnce({
          data: {...userProfile, info: {researchLabs: 'ru', otherAffiliations: []}}
        })
        await store.loadProfile()
        expect(store.profile.info.institution).toEqual(defaultUserInfo.institution)
        expect(store.profile.info.researchLabs).toEqual('ru')
      })
      it('should notify service error', testError(
          'get',
          t('notify.error.selfProfile'),
          store.loadProfile
      ))
    })
    describe('logIn', () => {
      describe('succes scenario', () => {
        const route = { path: '/' }
        beforeEach(async () => {
          mockedAxios.logIn.mockResolvedValueOnce()
          mockedRouter.replace.mockResolvedValueOnce()
          store.loadProfile = vi.fn().mockResolvedValueOnce({})
          await store.logIn(route, {username, password})
        })
        it('should call authority logIn method', () => {
          expect(mockedAxios.logIn).toHaveBeenCalledOnce()
        })
        it('should load user profile on success', () => {
          expect(store.loadProfile).toHaveBeenCalledOnce()
        })
        it('should provide credentials to logIn', () => {
          expect(mockedAxios.logIn).toHaveBeenCalledWith({username, password})
        })
        it('should route user out on success', () => {
          expect(mockedRouter.replace).toHaveBeenNthCalledWith(1, route)
        })
      })
      it('should notify service error', testError(
          'logIn',
          t('notify.error.login'),
          store.logIn
      ))
    })
    describe('logOut', () => {
      describe('without desired path', () => {
        const logoutUrl = `${import.meta.env.VITE_APP_AXIOS_BASE_URL}logout/`
        const logoutPattern = new RegExp(`^${logoutUrl}`)
        beforeEach(async () => {
          await store.logOut()
        })
        afterEach(async () => {
          mockedGetUrl.mockRestore()
          mockedGetUrl.mockRestore()
        })
        it('should call authority logOut method', () => {
          expect(mockedAxios.logOut).toHaveBeenCalledOnce()
        })
        it('should set a redirection url', () => {
          expect(mockedGetUrl).toHaveBeenCalledWith('')
        })
        it('should log user out', () => {
          expect(store.isLogged).toEqual(false)
        })
        it('should redirect user', () => {
          expect(mockedSetUrl).toHaveBeenCalledOnce()
        })
        it('should set window location to logout api', () => {
          expect(mockedSetUrl.mock.lastCall).toHaveLength(1)
          expect(mockedSetUrl.mock.lastCall?.[0]).toMatch(logoutPattern)
        })
      })
      it('should redirect to a given path', async () => {
        const desiredPath = 'my/custom/path'
        await store.logOut(desiredPath)
        expect(mockedGetUrl).toHaveBeenNthCalledWith(1, desiredPath)
      })
    })
    describe('updateProfile', () => {
      describe('success scenario', () => {
        const patchData = {lastName: 'was'}
        beforeEach(async () => {
          mockedAxios.patch.mockResolvedValueOnce({data: patchData})
          await store.updateProfile(patchData)
        })
        it('should should call user detail api',  () => {
          expect(mockedAxios.patch).toHaveBeenCalledOnce()
          expect(mockedAxios.patch).toHaveBeenLastCalledWith(`users/profile/`, patchData)
        })
        it('should set user profile with received data', () => {
          expect(store.profile).toEqual(patchData)
        })
        it('should dispatch a success notification', () => {
          testSuccess(t('profile.update.success'))
        })
      })
      it('should notify service error', testError(
          'patch',
          t('profile.update.failure'),
          store.updateProfile
      ))
    })
    describe('updateMail', () => {
      describe('success scenario', () => {
        const newMail = "new@mail.org"
        beforeEach(async () => {
          store.profile = userProfile
          mockedAxios.patch.mockResolvedValueOnce({data: {email: 'new@mail.org'}})
          await store.updateMail(newMail)
        })
        it('should call patch api with given email', () => {
          expect(mockedAxios.patch).toHaveBeenCalledOnce()
          expect(mockedAxios.patch).toHaveBeenLastCalledWith('users/profile/email/', {email: newMail})
        })
        it('should update profile with retrieve data', async () => {
          expect(store.email).toEqual(newMail)
        })
        it('should dispatch a success notification', () => {
          testSuccess(t('profile.update.success'))
        })
      })
      it('should notify service error', testError(
          'patch',
          t('profile.update.failure'),
          store.updateMail
      ))
    })
    describe('updatePassword', () => {
      const passwords =  {
        oldPassword: 'lambdabutstrong',
        newPassword: 'gammaandstronger',
        newPassword2: 'gammaandstronger'
      }
      it('should call patch api with given email', () => {
        mockedAxios.patch.mockResolvedValueOnce({data: userProfile})
        store.updatePassword(passwords)
        expect(mockedAxios.patch).toHaveBeenCalledOnce()
        expect(mockedAxios.patch).toHaveBeenLastCalledWith('users/profile/password/', passwords)
      })
    })
    describe('resetPassword', () => {
      it('should call post api with given email', () => {
        store.resetPassword(email)
        expect(mockedAxios.post).toHaveBeenCalledOnce()
        expect(mockedAxios.post).toHaveBeenLastCalledWith(
            'login/password-reset/',
            {url: `${appOrigin}/reset-password`, email}
        )
      })
      it('should notify mail successfully sent', async () => {
        mockedAxios.post.mockResolvedValueOnce({})
        await store.resetPassword(email)
        testSuccess(t('mail.success', {email}))
      })
      it('should notify mail sent failure', testError(
          'post',
          t('mail.failure', {email}),
          store.resetPassword, email
      ))
    })
  })
})
