import { describe, it, beforeEach, afterEach, expect, vi } from 'vitest'
import { createPinia, setActivePinia } from 'pinia'
import { useProjectStore } from '@/stores/project'
import { t } from '@/plugins/i18n'
import { authAxios } from '@/services/authService'
import { notifyError, notifySuccess } from '@/services/alertService'
import type { Mocked } from 'vitest'
import type { Authority } from '@/services/authService'
import { initProject } from '../../../src/stores/project'

vi.mock(
    '@/services/alertService',
    () => ({
      notifyError: vi.fn(),
      notifySuccess: vi.fn()
    })
)

vi.mock(
    '@/services/authService',
    () => ({
      authAxios: {
        delete: vi.fn(),
        get: vi.fn(),
        patch: vi.fn(),
        post: vi.fn(),
      }
    })
)
const mockedAxios: Mocked<Authority> = vi.mocked(authAxios, true)

describe('store/project', () => {
  const projectId = 'e34a-4f23'

  setActivePinia(createPinia())
  const store = useProjectStore()

  beforeEach(() => {
    store.$reset()
  })
  afterEach(() => {
    vi.restoreAllMocks()
  })
  describe('actions', () => {
    const axiosError = new Error('failure')
    const testError = (apiMethod: 'get' | 'patch' | 'post' | 'delete', message: string, storeAction: Function, ...params: any[]) =>
        async () => {
          mockedAxios[apiMethod].mockRejectedValueOnce(axiosError)
          await storeAction(...params)
          expect(notifyError).toHaveBeenCalledOnce()
          expect(notifyError).toHaveBeenLastCalledWith(
              message,
              axiosError
          )
        }
    const testSuccess = (message: string) => {
      expect(notifySuccess).toHaveBeenNthCalledWith(1, message)
    }
    beforeEach(() => {
      store.current = {
        ...initProject(),
        id: projectId,
        name: 'unnamed',
      }
    })
    describe('createProject', () => {
      const newProject = {
        name: 'project one',
        description: 'one of many',
        isPublic: false
      }
      it('should call post api with given project', () => {
        store.createProject(newProject)
        expect(mockedAxios.post).toHaveBeenCalledOnce()
        expect(mockedAxios.post).toHaveBeenLastCalledWith(
            'projects/',
            newProject
        )
      })
      it('should set work on project with retrieve data', async () => {
        const retrievedProject = {id:'e34a-4f23', ...newProject}
        mockedAxios.post.mockResolvedValueOnce({ data: retrievedProject})
        await store.createProject(newProject)
        expect(store.current).toEqual(retrievedProject)
      })
      it('should notify on failure', testError(
          'post',
          t('notify.error.project.create'),
          store.createProject, newProject
      ))
    })
    describe('loadProject', () => {
      const projectId = 'e34a-4f23'
      it('should call post api with given project', () => {
        store.loadProject(projectId)
        expect(mockedAxios.get).toHaveBeenNthCalledWith(1, 'projects/e34a-4f23/')
      })
      it('should set work on project with retrieve data', async () => {
        const getProject = {
          id: projectId,
          name: 'project one',
          description: 'one of many'
        }
        mockedAxios.get.mockResolvedValueOnce({ data: getProject})
        await store.loadProject(projectId)
        expect(store.current).toEqual(getProject)
      })
      it('should notify on failure', testError(
        'get',
          t('notify.error.project.load'),
          store.loadProject, projectId
      ))
    })
    describe('updateProject', () => {
      const patchData = {
        id : projectId,
        name: 'project one',
        description: 'still secret'
      }
      describe('success scenario', () => {
        const updatedProject = {...patchData}
        beforeEach(async () => {
          mockedAxios.patch.mockResolvedValueOnce({ data: updatedProject})
          await store.updateProject(patchData)
        })
        it('should call patch api of current project', () => {
          const {name, description} = patchData
          expect(mockedAxios.patch).toHaveBeenCalledOnce()
          expect(mockedAxios.patch).toHaveBeenLastCalledWith(
              'projects/e34a-4f23/',
              {name, description}
          )
        })
        it('should set work on project with retrieve data', () => {
          expect(store.current).toEqual(updatedProject)
        })
        it('should notify on success', () => {
          testSuccess(t('project.generalInfo.updated'))
        })
      })
      it('should notify on failure', testError(
          'patch',
          t('notify.error.update'),
          store.updateProject, patchData
      ))
    })
    describe('retrieveProjects', () => {
        it('should call retrieve list api', () => {
          store.retrieveProjects(1, 11)
          expect(mockedAxios.get).toHaveBeenCalledOnce()
          expect(mockedAxios.get).toHaveBeenLastCalledWith(
              'projects/?page=1&page_size=11'
          )
        })
        it('should set contribution list with retrieve data', async () => {
          const projectList = [
            {id: 'e321-47bd', name: 'project one', isManager: true},
            {id: 'e321-47bd', name: 'project two', isManager: false},
            {id: 'e321-47bd', name: 'project four', isManager: false},
          ]
          mockedAxios.get.mockResolvedValueOnce({
            data: {count: 3, results: projectList}
          })
          await store.retrieveProjects(1, 10)
          expect(store.contributions).toEqual(projectList)
        })
        it('should set pageCount list with retrieve data', async () => {
          mockedAxios.get.mockResolvedValueOnce({
            data: {count: 38, nextPage: 2, results: []}
          })
          await store.retrieveProjects(1, 10)
          expect(store.pageCount).to.equal(4, 'basic ration count')
        })
        it('should notify on failure', testError(
            'get',
            t('notify.error.project.all'),
            store.retrieveProjects,
        ))
      })
    describe('requestDeletion', () => {
      const projectIds = ['3425-43e4', '654f-547b']
      it('should call delete api', () => {
        mockedAxios.delete.mockResolvedValueOnce({})
        store.requestDeletion(projectIds)
        expect(mockedAxios.delete).toHaveBeenCalledOnce()
        expect(mockedAxios.delete).toHaveBeenLastCalledWith(
            'projects/request_deletion/',
            {data: {projectIds}}
        )
      })
      it('should notify that the request has been sent', async () => {
        mockedAxios.delete.mockResolvedValueOnce({})
        await store.requestDeletion(projectIds)
        testSuccess(t('notify.delete.request.success'))
      })
      it('should notify explicit error', testError(
          'delete',
          t('notify.delete.request.error'),
          store.requestDeletion, []
      ))
    })
  })
})
